Fixed get_xattr_acl() -- it needed to zero *len_p.
[rsync/rsync.git] / xattrs.c
index f662cf7..2345214 100644 (file)
--- a/xattrs.c
+++ b/xattrs.c
@@ -6,8 +6,9 @@
  * Copyright (C) 2006, 2007 Wayne Davison
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License version 2 as
- * published by the Free Software Foundation.
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 3 of the License, or
+ * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
- * 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
+ * with this program; if not, visit the http://fsf.org website.
  */
 
 #include "rsync.h"
+#include "ifuncs.h"
 #include "lib/sysxattrs.h"
 
 #ifdef SUPPORT_XATTRS
@@ -30,6 +31,7 @@ extern int am_sender;
 extern int am_generator;
 extern int read_only;
 extern int list_only;
+extern int preserve_xattrs;
 extern int checksum_seed;
 
 #define RSYNC_XAL_INITIAL 5
@@ -62,7 +64,8 @@ extern int checksum_seed;
 #define RPRE_LEN ((int)sizeof RSYNC_PREFIX - 1)
 
 #define XSTAT_ATTR RSYNC_PREFIX "%stat"
-#define XSTAT_LEN ((int)sizeof XSTAT_ATTR - 1)
+#define XACC_ACL_ATTR RSYNC_PREFIX "%aacl"
+#define XDEF_ACL_ATTR RSYNC_PREFIX "%dacl"
 
 typedef struct {
        char *datum, *name;
@@ -89,7 +92,7 @@ static void rsync_xal_free(item_list *xalp)
        xalp->count = 0;
 }
 
-void free_xattr(statx *sxp)
+void free_xattr(stat_x *sxp)
 {
        if (!sxp->xattr)
                return;
@@ -223,8 +226,9 @@ static int rsync_xal_get(const char *fname, item_list *xalp)
                        continue;
 #endif
 
-               if (am_root < 0 && name_len == XSTAT_LEN + 1
-                && name[RPRE_LEN] == '%' && strcmp(name, XSTAT_ATTR) == 0)
+               /* No rsync.%FOO attributes are copied w/o 2 -X options. */
+               if (preserve_xattrs < 2 && name_len > RPRE_LEN
+                && name[RPRE_LEN] == '%' && HAS_PREFIX(name, RSYNC_PREFIX))
                        continue;
 
                datum_len = name_len; /* Pass extra size to get_xattr_data() */
@@ -266,7 +270,7 @@ static int rsync_xal_get(const char *fname, item_list *xalp)
 }
 
 /* Read the xattr(s) for this filename. */
-int get_xattr(const char *fname, statx *sxp)
+int get_xattr(const char *fname, stat_x *sxp)
 {
        sxp->xattr = new(item_list);
        *sxp->xattr = empty_xattr;
@@ -328,7 +332,7 @@ static void rsync_xal_store(item_list *xalp)
 }
 
 /* Send the make_xattr()-generated xattr list for this flist entry. */
-int send_xattr(statx *sxp, int f)
+int send_xattr(stat_x *sxp, int f)
 {
        int ndx = find_matching_xattr(sxp->xattr);
 
@@ -374,7 +378,7 @@ int send_xattr(statx *sxp, int f)
 /* Return a flag indicating if we need to change a file's xattrs.  If
  * "find_all" is specified, also mark any abbreviated xattrs that we
  * need so that send_xattr_request() can tell the sender about them. */
-int xattr_diff(struct file_struct *file, statx *sxp, int find_all)
+int xattr_diff(struct file_struct *file, stat_x *sxp, int find_all)
 {
        item_list *lst = rsync_xal_l.items;
        rsync_xa *snd_rxa, *rec_rxa;
@@ -636,8 +640,9 @@ void receive_xattr(struct file_struct *file, int f)
                        continue;
                }
 #endif
-               if (am_root < 0 && name_len == XSTAT_LEN + 1
-                && name[RPRE_LEN] == '%' && strcmp(name, XSTAT_ATTR) == 0) {
+               /* No rsync.%FOO attributes are copied w/o 2 -X options. */
+               if (preserve_xattrs < 2 && name_len > RPRE_LEN
+                && name[RPRE_LEN] == '%' && HAS_PREFIX(name, RSYNC_PREFIX)) {
                        free(ptr);
                        continue;
                }
@@ -654,9 +659,9 @@ void receive_xattr(struct file_struct *file, int f)
        F_XATTR(file) = ndx;
 }
 
/* Turn the xattr data in statx into cached xattr data, setting the index
  * values in the file struct. */
-void cache_xattr(struct file_struct *file, statx *sxp)
+void cache_xattr(struct file_struct *file, stat_x *sxp)
 {
        int ndx;
 
@@ -671,7 +676,7 @@ void cache_xattr(struct file_struct *file, statx *sxp)
 }
 
 static int rsync_xal_set(const char *fname, item_list *xalp,
-                        const char *fnamecmp, statx *sxp)
+                        const char *fnamecmp, stat_x *sxp)
 {
        rsync_xa *rxas = xalp->items;
        ssize_t list_len;
@@ -778,7 +783,7 @@ static int rsync_xal_set(const char *fname, item_list *xalp,
 
 /* Set extended attributes on indicated filename. */
 int set_xattr(const char *fname, const struct file_struct *file,
-             const char *fnamecmp, statx *sxp)
+             const char *fnamecmp, stat_x *sxp)
 {
        int ndx;
        item_list *lst = rsync_xal_l.items;
@@ -795,6 +800,32 @@ int set_xattr(const char *fname, const struct file_struct *file,
        return rsync_xal_set(fname, lst + ndx, fnamecmp, sxp);
 }
 
+#ifdef SUPPORT_ACLS
+char *get_xattr_acl(const char *fname, int is_access_acl, size_t *len_p)
+{
+       const char *name = is_access_acl ? XACC_ACL_ATTR : XDEF_ACL_ATTR;
+       *len_p = 0; /* no extra data alloc needed from get_xattr_data() */
+       return get_xattr_data(fname, name, len_p, 1);
+}
+
+int set_xattr_acl(const char *fname, int is_access_acl, const char *buf, size_t buf_len)
+{
+       const char *name = is_access_acl ? XACC_ACL_ATTR : XDEF_ACL_ATTR;
+       if (sys_lsetxattr(fname, name, buf, buf_len) < 0) {
+               rsyserr(FERROR, errno,
+                       "set_xattr_acl: lsetxattr(\"%s\",\"%s\") failed",
+                       fname, name);
+               return -1;
+       }
+       return 0;
+}
+
+int del_def_xattr_acl(const char *fname)
+{
+       return sys_lremovexattr(fname, XDEF_ACL_ATTR);
+}
+#endif
+
 int get_stat_xattr(const char *fname, int fd, STRUCT_STAT *fst, STRUCT_STAT *xst)
 {
        int mode, rdev_major, rdev_minor, uid, gid, len;
@@ -886,7 +917,7 @@ int set_stat_xattr(const char *fname, struct file_struct *file)
                fst.st_rdev = 0; /* just in case */
 
        if (mode == fmode && fst.st_rdev == rdev
-        && fst.st_uid == F_UID(file) && fst.st_gid == F_GID(file)) {
+        && fst.st_uid == F_OWNER(file) && fst.st_gid == F_GROUP(file)) {
                /* xst.st_mode will be 0 if there's no current stat xattr */
                if (xst.st_mode && sys_lremovexattr(fname, XSTAT_ATTR) < 0) {
                        rsyserr(FERROR, errno,
@@ -898,12 +929,12 @@ int set_stat_xattr(const char *fname, struct file_struct *file)
        }
 
        if (xst.st_mode != fmode || xst.st_rdev != rdev
-        || xst.st_uid != F_UID(file) || xst.st_gid != F_GID(file)) {
+        || xst.st_uid != F_OWNER(file) || xst.st_gid != F_GROUP(file)) {
                char buf[256];
                int len = snprintf(buf, sizeof buf, "%o %u,%u %u:%u",
                        to_wire_mode(fmode),
                        (int)major(rdev), (int)minor(rdev),
-                       (int)F_UID(file), (int)F_GID(file));
+                       F_OWNER(file), F_GROUP(file));
                if (sys_lsetxattr(fname, XSTAT_ATTR, buf, len) < 0) {
                        if (errno == EPERM && S_ISLNK(fst.st_mode))
                                return 0;
@@ -654,9 +659,9 @@ void receive_xattr(struct file_struct *file, int f)
 }
 
-/* Turn the xattr data in statx into cached xattr data, setting the index
+/* Turn the xattr data in stat_x into cached xattr data, setting the index
  * values in the file struct. */